import { Injectable } from "@nestjs/common";
import { Cron, CronExpression } from "@nestjs/schedule";
import { InjectRepository } from "@nestjs/typeorm";
import {
    fetchAllProductCategory,
    fetchAllProductsForScheduler,
    fetchAllRetailers,
    fetchProductsCount
} from "src/helper/dutchie-plus-query";
import { getGraphQLClient } from "src/helper/graphql-client-helper";
import { CreateProductCategoryDto } from "../dutchie-plus/custom/product-category/dto/create-product-category.dto";
import { ProductCategoryRepository } from "../dutchie-plus/custom/product-category/product-category.repository";
import { CreateProductServiceDto } from "../dutchie-plus/custom/product-service/dto/create-product-service.dto";
import { ProductServiceRepository } from "../dutchie-plus/custom/product-service/product-service.repository";
import { urlSlug } from "src/helper/common-helper";

@Injectable()
export class SchedulerService {
    constructor(
        @InjectRepository(ProductCategoryRepository)
        private readonly productCatRepo: ProductCategoryRepository,
        @InjectRepository(ProductServiceRepository)
        private readonly productServiceRepo: ProductServiceRepository
    ) {
    }

    @Cron(CronExpression.EVERY_HOUR)
    async processProductData(productData: any, retailerId: string, retailerName: string): Promise<any | Error> {
        try {
            const [
                productService,
                productCategory
            ] = await Promise.all([
                this.productServiceRepo.findByFilterQuery({}),
                this.productCatRepo.findAll()
            ]);

            const productServiceDatasetIds = new Set(productService.map((item) => item.productId));
            const productMapService: any = new Map(productService.map(item => [item.productId, item]));
            const productDifferences = [];
            const differeneceDeleteIds = [];
            const categoriesArr = [];

            await Promise.all(productData.menu.products.map(async (item) => {
                // product service
                if (productServiceDatasetIds.size > 0) {
                    if (!productServiceDatasetIds.has(item.id)) {
                        productDifferences.push(item);
                    } else {
                        // Now you can find a product by id in constant time:
                        const matchingItem = productMapService.get(item.id);
                        // const matchingItem = productService.find((item2) => item2.productId === item.id);
                        const brand = item.brand !== null ? item.brand.name : "";
                        if (
                            matchingItem &&
                            (matchingItem.brandName !== brand ||
                                matchingItem.image !== item.image ||
                                matchingItem.productName !== item.name ||
                                matchingItem.productSlug !== item.slug ||
                                matchingItem.category !== item.category ||
                                matchingItem.subCategory !== item.subcategory ||
                                matchingItem.strainType !== item.strainType)
                        ) {
                            differeneceDeleteIds.push(item.id);
                            productDifferences.push(item);
                        }
                    }
                } else {
                    productDifferences.push(item);
                }

                // all categories
                categoriesArr.push(item.category);
            }));


            // if any update occers delete
            if (differeneceDeleteIds.length > 0) {
                await this.productServiceRepo.deleteRowsFromDatabase(differeneceDeleteIds);
            }

            // create entities
            console.log(retailerName, ">", " New Product = ", productDifferences.length);
            const createProductServiceDtos = productDifferences.map((item) => {
                const brand = item.brand !== null ? item.brand.name : "";
                return {
                    retailerId,
                    retailerName,
                    productId: item.id,
                    productName: item.name,
                    productSlug: item.slug,
                    image: item.image,
                    brandName: brand,
                    category: item.category,
                    subCategory: item.subcategory,
                    strainType: item.strainType,
                    variants: item.variants.length
                };
            });
            
            if (createProductServiceDtos.length > 0) {
                await this.productServiceRepo.createEntities(createProductServiceDtos);
            }

            // check category
            const categories = Array.from(new Set(categoriesArr));

            if (Object.keys(productCategory).length > 0) {
                const difference = categories.filter((category) =>
                    !productCategory.some((c) => c.categoryName === category)
                );
                console.log("Product Categories Difference = ", difference);
                // if difference found then create entities with difference product category
                const createProductCategoryDtos = difference.map((item) => {
                    return {
                        categoryName: item,
                        urlSlug: urlSlug(item)
                    };
                });

                if (createProductCategoryDtos.length > 0) {
                    await this.productCatRepo.createEntities(createProductCategoryDtos);
                }
            } else {
                // if product categories table is empty in our local database then create all the category in local
                const createProductCategoryDtos = categories.map((item) => {
                    return {
                        categoryName: item,
                        urlSlug: urlSlug(item)
                    };
                });

                if (createProductCategoryDtos.length > 0) {
                    await this.productCatRepo.createEntities(createProductCategoryDtos);
                }
            }

        } catch (e) {
            console.log(e);
            return e;
        }
    }


    @Cron(CronExpression.EVERY_HOUR)
    async createUpdateDutchiePlusProductsAndCategory() {
        try {
            console.log(">>>Run Create or Update Dutchie Plus Products Scheduler<<<");

            // Get GraphQL client
            const graphQLClient = getGraphQLClient();

            // Fetch all retailers
            const query = fetchAllRetailers();
            const data = await graphQLClient.request(query);
            const retailers = data.retailers;

            // Check if retailers exist
            if (retailers.length > 0) {
                // Create an array of promises for fetching product data
                console.log("Start Fetching product count Parallel  =", new Date());
                const fetchProductsCountObject = fetchProductsCount();
                const retailerListPromises = retailers.map((retailer) =>
                    graphQLClient.request(fetchProductsCountObject, { retailerId: retailer.id })
                        .then(response => {
                            if (response && response.menu && response.menu.productsCount != null) {

                                return {
                                    retailerId: retailer.id,
                                    retailerName: retailer.name,
                                    pagination: {
                                        offset: 0,
                                        limit: response.menu.productsCount
                                    }
                                };
                            }
                        })
                );

                let allRetailerProductCount = await Promise.all(retailerListPromises);
                allRetailerProductCount = allRetailerProductCount.filter((data) => data.pagination.limit > 0);
                console.log("End Fetching product count Parallel =", new Date());


                console.log("Start Fetching product Parallel End =", new Date());
                const fetchProductsQuery = fetchAllProductsForScheduler();
                const fetchProductPromises = allRetailerProductCount.map(variable =>
                    graphQLClient.request(fetchProductsQuery, variable).then(response => {
                        if (response && response.menu && response.menu.products) {
                            return {
                                retailerId: variable.retailerId,
                                retailerName: variable.retailerName,
                                productData: response
                            };
                        }
                    })
                );
                console.log("End Fetching product Parallel End =", new Date());

                const allProductData = await Promise.all(fetchProductPromises);
                //console.log(allProductData)
                for (const data of allProductData) {
                    console.log("Send for Processing", new Date());
                    await this.processProductData(data.productData, data.retailerId, data.retailerName);
                    console.log("End processing =", new Date());
                }
                console.log(">>>End Create or Update Dutchie Plus Products Scheduler<<<");
            } else {
                console.log("Retailers Not Found!");
            }
        } catch (e) {
            console.log(e);
            return e;
        }
    }
}
